#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/time.h>
#include <unistd.h>
#include <ncurses.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <limits.h>
#include <time.h>
#include <syslog.h>
#include <string.h>

extern char *optarg;

#include <sndfile.h>

#include "error.h"
#include "snd_dev.h"
#include "utils.h"

#define SAMPLE_RATE 	44100

#ifndef WAV_PATH
#define WAV_PATH	"listen_wav_out"
#endif /* WAV_PATH */

#ifndef DSP_PATH
#define DSP_PATH	"/dev/dsp"
#endif /* DSP_PATH */

#ifndef CONFIGFILE
#define CONFIGFILE	"/usr/local/etc/listener.conf"
#endif /* CONFIGFILE */


char *wav_path = NULL; 					/* path for wav-files */
char *dev_name = NULL; 					/* path to audio-device */
char *configfile = NULL;				/* configurationfile */

int detect_level = 754;					/* threshold for sound-detection */
int rec_silence = 2;					/* number of seconds to keep on reading when silence is detected */
int min_duration = 1;					/* how many samples above threshold before recording starts */
int sample_rate = SAMPLE_RATE;				/* speaks for itself */
int compression = SF_FORMAT_PCM_16;			/* default is no compression */
char *exec = NULL;					/* what to execute after recording */

char do_exit = 0;

int get_compression_type( const char *pcCompression )
{
	if (strcasecmp(pcCompression, "u-law") == 0)
		return SF_FORMAT_ULAW;
	else if (strcasecmp(pcCompression, "a-law") == 0)
		return SF_FORMAT_ALAW;
	else if (strcasecmp(pcCompression, "IMA-ADPCM") == 0)
		return SF_FORMAT_IMA_ADPCM;
	else if (strcasecmp(pcCompression, "MS-ADPCM") == 0)
		return SF_FORMAT_MS_ADPCM;
	else if (strcasecmp(pcCompression, "GSM-6.10") == 0)
		return SF_FORMAT_GSM610;
	else
		return -1;
} /* int get_compression_type( ) */

void start_recording(int fd, short *first, int first_n_samples)
{
	short *buffer;
	SNDFILE *sndfile_out;
	SF_INFO sfinfo_out;
	char fname[PATH_MAX];
	struct tm *tmdummy;
	time_t recstart, prev, now;
	pid_t pid;

	buffer = (short *)mymalloc(sample_rate * sizeof(short), "recording buffer(2)");

	/* create filename */
	time(&recstart);
	tmdummy = localtime(&recstart);
	sprintf(fname, "%s/%04d-%02d-%02d_%02d%02d%02d.wav",
		wav_path,
		tmdummy -> tm_year + 1900,
		tmdummy -> tm_mon + 1,
		tmdummy -> tm_mday,
		tmdummy -> tm_hour,
		tmdummy -> tm_min,
		tmdummy -> tm_sec);
	prev = now = recstart;

	/* create file */
	memset(&sfinfo_out, 0x00, sizeof(sfinfo_out));
        sfinfo_out.samplerate = sample_rate;
        sfinfo_out.channels = 1;
        sfinfo_out.format = SF_FORMAT_WAV | compression;
        sndfile_out = sf_open(fname, SFM_WRITE, &sfinfo_out);
	if (!sndfile_out)
		error_exit("Could not create file %s!", fname);

	syslog(LOG_INFO, "Started recording to %s", fname);

	/* write sample to disk */
	if (sf_writef_short(sndfile_out, first, first_n_samples) != first_n_samples)
		error_exit("failed to write to wav-file\n");

	for(; (now - prev) < rec_silence && do_exit == 0; )
	{
		int loop, threshold_count = 0;

		/* read sample */
		if (READ(fd, (char *)buffer, sample_rate * 2 * 1) != (sample_rate * 2 * 1))
			error_exit("failed to read from audio-device\n");

		/* write sample to disk */
		if (sf_writef_short(sndfile_out, buffer, sample_rate) != sample_rate)
			error_exit("failed to write to wav-file\n");

		/* see if it became silent again */
		for(loop=0; loop<sample_rate; loop++)
		{
			if (abs(buffer[loop]) > detect_level)
			{
				threshold_count++;

				if (threshold_count >= min_duration)
				{
					time(&prev);
					break;
				}
			}
		}

		time(&now);
	}

	sf_close(sndfile_out);

	free(buffer);

	syslog(LOG_INFO, "Stopped recording to %s (%d seconds)", fname, (int)(now - recstart));

	if (exec)
	{
		syslog(LOG_INFO, "Starting childprocess: %s", exec);

		pid = fork();

		if (pid == -1)
		{
			syslog(LOG_ERR, "Failed to fork! %m");
		}
		else if (pid == 0)
		{
			if (-1 == execlp(exec, exec, fname, (void *)NULL))
				error_exit("Failed to start childprocess %s with parameter %s: %m", exec, fname);
		}
	}
}

void sigh(int sig)
{
	if (sig == SIGTERM)
		do_exit = 1;
}


void prt_usage( const char *pcProgram )
{
	printf( "\n"
		"Usage: %s [options]\n"
		"-C<compression>  Set WAV compression     -e<command>      Script to call after recording\n"
		"-r<rate>         Sample rate             -m<min_duration> Minimum duration to record\n"
		"-R<rec_silence>  Silence to record       -l<detect_level> Detect level\n"
		"-c<configfile>   Configfile to use\n"
		"-w<wave-dir>     Where to write .WAVs    -d<device>       DSP device to use (/dev/dsp)\n"
		"-s               Be silent               -h               This help text\n\n",
			pcProgram );
}

/* This specifies the minimum user-supplied settings we need without a 
 * specified configuration-file.
 */
#define MAX_SETTINGS	6
int main(int argc, char *argv[])
{
	int	iUserSet = 0;	/* user supplied settings counter */
	int	iSwitch;	/* for processing command-line arguments */
	int	iShutup = 0;	/* be silent on stdout? */
	char	show_help = 0;

	int fd;
	FILE *fh;
	/* right now I'm not sure if a short is always 2 bytes. check this on your
	 * platform
	 */
	short *buffer;
	int buffer_size;

	/* Check command-line options;
	 */
	while ( (iSwitch = getopt( argc, argv, "c:shS::r:R:l:m:C:e:w:d:" )) != -1 )
	{
		switch (iSwitch)
		{
		  case 'w':	/* WAV directory */
			wav_path = strdup(optarg);
			iUserSet++;
		 	break;

		  case 'd':	/* Device */
			dev_name = strdup(optarg);
			iUserSet++;
			break;

		  case 'e':	/* Exec */
			exec = strdup( optarg );
			break;

		  case 'C':	/* Compression */
			if ((compression = get_compression_type( optarg )) == -1)
			{
				printf( "Invalid compression type specified!\n" );
				return 2;
			}
			iUserSet++;
			break;

		  case 'm':	/* Min duration */
			min_duration = atoi( optarg );
			iUserSet++;
			break;

		  case 'l':	/* Detect level */
			detect_level = atoi( optarg );
			iUserSet++;
			break;

		  case 'R':	/* Record silence */
			rec_silence = atoi( optarg );
			iUserSet++;
			break;

		  case 'r':	/* Sample rate */
			sample_rate = atoi( optarg );
			iUserSet++;
			break;
			
		  case 'h':	/* Help -- display usage */
			show_help = 1;
			break;

		  case 'c':	/* Configuration-file */
			configfile = strdup(optarg);
			break;

		  case 's':	/* Be silent.. */
			iShutup = 1;
			break;

		  default:
			/* unknown switch */
			show_help = 1;
			break;
		} /* switch.. */	

	} /* while.. */

	if (iShutup == 0)
		printf("listener v" VERSION ", (C)2003-2004 by folkert@vanheusden.com\n");

	/* determine device to open */
	if ((iUserSet > 0) && (iUserSet < MAX_SETTINGS))
		show_help = 1;

	if (show_help == 1)
	{
		prt_usage(argv[0]);	
		return 1;
	}
	
	/* Load configfile if necessary... */
	if (iUserSet < MAX_SETTINGS)
	{
		/* Apply default settings;
		 */
		if (configfile == NULL) configfile = strdup( CONFIGFILE );
		if (dev_name == NULL) dev_name = strdup( DSP_PATH );
		if (wav_path == NULL) wav_path = strdup( WAV_PATH );

		fh = fopen(configfile, "rb");

		
		if (!fh) 
			error_exit("error opening configfile %s\n", configfile);

		/* Free up the memory allocated by the strdup( ) if that has been allocated;
		 */
		if (configfile)
			free(configfile);

		for(;;)
		{
			char *i_is, *cmd, *par = NULL;
			char *line = read_line(fh);
			if (!line)
				break;

	                // find '='
	                i_is = strchr(line, '=');
	                if (i_is)
	                {
	                        par = get_token(i_is+1);
	                        *i_is = 0x00;
	                }
	                cmd = get_token(line);
	
			if (strcasecmp(cmd, "wav_path") == 0)
				wav_path = strdup(par);
			else if (strcasecmp(cmd, "devname") == 0)
				dev_name = strdup(par);
			else if (strcasecmp(cmd, "detect_level") == 0)
				detect_level = atoi(par);
			else if (strcasecmp(cmd, "rec_silence") == 0)
				rec_silence = atoi(par);
			else if (strcasecmp(cmd, "min_duration") == 0)
				min_duration = atoi(par);
			else if (strcasecmp(cmd, "sample_rate") == 0)
				sample_rate = atoi(par);
			else if (strcasecmp(cmd, "compression") == 0)
			{
				if ((compression = get_compression_type( par )) == -1)
					error_exit("%s is an unknown compressionformat\n");
			}
			else if (strcasecmp(cmd, "exec") == 0)
				exec = strdup(par);
			else if (cmd[0] != '#')
				error_exit("'%s=%s' is not a known configuration-statement\n", cmd?cmd:"(none)", par?par:"(none)");

			free(par);
			free(cmd);
			free(line);
		}
		fclose(fh);
	}

	if (min_duration < 1)
		error_exit("min_duration must be at least 1!");

	/* open audio-device */
	fd = open(dev_name, O_RDWR);
	if (fd == -1)
		error_exit("Error opening device %s\n", dev_name);

	/* set audio-device to 44.1KHz 16bit mono */
	set_soundcard_sample_format(fd);
	set_soundcard_number_of_channels(fd, 1);
	set_soundcard_sample_rate(fd, sample_rate);

	if (iShutup == 0)
	{
		printf("Path:         %s\n", wav_path);
		printf("Device:       %s\n", dev_name);
		printf("Level:        %d\n", detect_level);
		printf("Rec silence:  %d\n", rec_silence);
		printf("Min duration: %d\n", min_duration);
		printf("Samplerate:   %d\n", sample_rate);
	} /* if(iShutup.. */

	if (daemon(-1, -1) == -1)
		error_exit("problem becoming daemon-process\n");

	syslog(LOG_INFO, "listener started");

	buffer_size = sample_rate * sizeof(short);
	buffer = (short *)mymalloc(buffer_size, "recording buffer(1)");

	signal(SIGTERM, sigh);
	signal(SIGCHLD, SIG_IGN);

	for(;do_exit == 0;)
	{
		int threshold_count = 0;
		int rc, loop;

		/* check audio-level */
		/* read data */
		rc = READ(fd, (char *)buffer, buffer_size);
		if (rc < 0)
			error_exit("error while reading from audio-device!\n", rc);

		/* check audio-levels */
		for(loop=0; loop<sample_rate; loop++)
		{
			if (abs(buffer[loop]) > detect_level)
			{
				threshold_count++;

				if (threshold_count >= min_duration)
				{
					start_recording(fd, buffer, sample_rate);
					break;
				}
			}
		}
	}

	/* Also free these strdup( )'d variables;
	 */
	if (exec) free( exec );
	if (dev_name) free( dev_name );
	if (wav_path) free( wav_path );

	return 0;
}
